from __future__ import annotations

import dataclasses
from dataclasses import dataclass
from typing import Any, Dict, Optional, Union

from pytest_sosu.utils import str_or_none


@dataclass(frozen=True)
class BaseBrowser:
    name: str
    version: str = "latest"


@dataclass(frozen=True)
class Browser(BaseBrowser):
    @classmethod
    def default(cls) -> Optional[Browser]:
        return cls(name="chrome")

    def __init__(self, name: str, version: Union[str, int] = "latest") -> None:
        super().__init__(name=name, version=str(version))

    @property
    def full_name(self) -> str:
        return f"{self.name} {self.version}"

    @property
    def slug(self) -> str:
        return f"{self.name}-{self.version}"

    def __str__(self) -> str:
        return self.full_name

    def __structlog__(self) -> Dict[str, Any]:
        return self.to_dict()

    def to_dict(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)


@dataclass(frozen=True)
class BasePlatform:
    name: str
    version: Optional[str] = None


@dataclass(frozen=True)
class Platform(BasePlatform):
    @classmethod
    def default(cls) -> Optional[Platform]:
        return None

    @classmethod
    def windows_default(cls) -> Optional[Platform]:
        return cls(name="Windows")

    def __init__(self, name: str, version: Optional[Union[str, int]] = None) -> None:
        super().__init__(name=name, version=str_or_none(version))

    @property
    def full_name(self) -> str:
        if self.version is None:
            return f"{self.name}"
        return f"{self.name} {self.version}"

    @property
    def slug(self) -> str:
        if self.version is None:
            return f"{self.name}"
        return f"{self.name}-{self.version}"

    def __str__(self) -> str:
        return self.full_name

    def __structlog__(self) -> Dict[str, Any]:
        return self.to_dict()

    def to_dict(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional
from urllib.parse import urlparse


@dataclass(frozen=True)
class WebDriverUrlData:
    host: str
    port: Optional[int] = None
    scheme: str = "https"
    username: Optional[str] = None
    access_key: Optional[str] = None
    path: str = "/wd/hub"

    @classmethod
    def from_url(cls, url: str) -> WebDriverUrlData:
        result = urlparse(url)
        if result.hostname is None:
            raise ValueError("missing hostname")
        return cls(
            scheme=result.scheme,
            username=result.username,
            access_key=result.password,
            host=result.hostname,
            port=result.port,
            path=result.path,
        )

    @property
    def address(self) -> str:
        if self.port is None:
            return self.host
        return f"{self.host}:{self.port}"

    @property
    def _access_key_starred(self) -> Optional[str]:
        if self.access_key is None:
            return None
        return "*" * len(self.access_key)

    @property
    def has_credentials(self) -> bool:
        return bool(self.username and self.access_key)

    def __repr__(self) -> str:
        url = self.to_safe_url()
        return f"{self.__class__.__name__}({url!r})"

    def to_url(self) -> str:
        return self._to_url(self.access_key)

    def to_safe_url(self) -> str:
        return self._to_url(self._access_key_starred)

    def _to_url(self, access_key: Optional[str]) -> str:
        auth_prefix = ""
        if self.username and access_key:
            auth_prefix = f"{self.username}:{access_key}@"
        return f"{self.scheme}://{auth_prefix}{self.address}{self.path}"  # noqa: E501

    def with_credentials(self, username, access_key) -> WebDriverUrlData:
        return self.__class__(
            host=self.host,
            port=self.port,
            scheme=self.scheme,
            username=username,
            access_key=access_key,
            path=self.path,
        )



from __future__ import annotations

from pytest_sosu.exceptions import (  # noqa: F401
    WebDriverTestFailed,
    WebDriverTestInterrupted,
)
from pytest_sosu.webdriver.capabilities import Capabilities  # noqa: F401
from pytest_sosu.webdriver.capabilities import CapabilitiesMatrix  # noqa: F401
from pytest_sosu.webdriver.capabilities import SauceOptions  # noqa: F401
from pytest_sosu.webdriver.driver import remote_webdriver_ctx  # noqa: F401
from pytest_sosu.webdriver.platforms import Browser, Platform  # noqa: F401
from pytest_sosu.webdriver.url import WebDriverUrlData  # noqa: F401



import logging

import structlog
from structlog.stdlib import filter_by_level, render_to_log_kwargs
from structlog.types import EventDict


def get_struct_logger(name):
    return _get_stdlib_struct_logger(name)


def _get_stdlib_struct_logger(name):
    logger = structlog.wrap_logger(
        logging.getLogger(name),
        processors=[
            filter_by_level,
            render_event_dict_into_msg,
            render_to_log_kwargs,
        ],
    )
    return logger


def render_event_dict_into_msg(
    logger: logging.Logger, method_name: str, event_dict: EventDict
):
    event = event_dict.pop("event")
    msg = event + "     " + " ".join(_iter_kv_strings(event_dict))
    return {"event": msg}


def _iter_kv_strings(event_dict: EventDict):
    for name, value in event_dict.items():
        yield f"{name}={value}"



from typing import Any, List, Optional

import pytest
from _pytest.mark.structures import Mark
from _pytest.python import Metafunc

from pytest_sosu.exceptions import (
    InvalidMarkerConfiguration,
    MultipleMarkerParametersFound,
)
from pytest_sosu.webdriver import Capabilities, CapabilitiesMatrix

SOSU_MARKER_NAME = "sosu"


def parametrize_capabilities(metafunc: Metafunc):
    caps_list = get_marker_capabilites_list_or_none(metafunc)
    if (
        "sosu_webdriver_parameter_capabilities" in metafunc.fixturenames
        and caps_list is not None
    ):
        metafunc.parametrize(
            "sosu_webdriver_parameter_capabilities",
            [pytest.param(c, id=c.slug) for c in caps_list],
        )


def get_marker_capabilites_list_or_none(
    metafunc: Metafunc,
) -> Optional[List[Capabilities]]:
    markers = get_sosu_markers(metafunc)
    caps = _get_marker_capabilites_list_or_none(markers)
    caps_matrix = _get_marker_capabilites_matrix_or_none(markers)

    if caps is not None and caps_matrix is not None:
        raise InvalidMarkerConfiguration(
            "both capabilities_matrix and capabilities are provided"
        )

    if caps is not None:
        return caps

    if caps_matrix is not None:
        return list(caps_matrix.iter_capabilities())

    return None


def get_sosu_markers(metafunc: Metafunc) -> List[Mark]:
    return [
        item
        for item in metafunc.definition.own_markers
        if item.name == SOSU_MARKER_NAME
    ]


def _get_marker_capabilites_matrix_or_none(
    sosu_markers: List[Mark],
) -> Optional[CapabilitiesMatrix]:
    return get_marker_unique_parameter_or_none(sosu_markers, "capabilities_matrix")


def _get_marker_capabilites_list_or_none(
    sosu_markers: List[Mark],
) -> Optional[List[Capabilities]]:
    caps = get_marker_unique_parameter_or_none(sosu_markers, "capabilities")
    if isinstance(caps, Capabilities):
        return [caps]
    return caps


def get_marker_unique_parameter_or_none(sosu_markers: List[Mark], key: str) -> Any:
    values = [m.kwargs.get(key) for m in sosu_markers]
    values = [v for v in values if v is not None]

    if len(values) > 1:
        raise MultipleMarkerParametersFound()

    if not values:
        return None

    return values[0]
